import os, re
from jlp import logger, get_config_option
import jenkinsutils
from . import get_json_jenkins

RE_LP_BUG_IDS = r'LP:(,? #\d+)+\s*$'


class LaunchpadVote():
    APPROVE = 'Approve'
    DISAPPROVE = 'Disapprove'
    NEEDS_FIXING = 'Needs Fixing'

def get_source_repo(mp):
    if '+git' in mp.web_link:
        return mp.source_git_repository.git_identity
    else:
        return mp.source_branch.bzr_identity

def get_source_branch(mp):
    if '+git' in mp.web_link:
        return mp.source_git_path.replace('refs/heads/', '')
    else:
        return mp.source_branch.bzr_identity

def get_target_branch(mp):
    if '+git' in mp.web_link:
        return mp.target_git_path.replace('refs/heads/', '')
    else:
        return mp.target_branch.bzr_identity

def get_vote_subject(mp):
    """Given a mp handle return a subject for the vote message

    Unfortunately there is no method in the API that gives you the "standard"
    subject that launchapd is using and some email clients (gmail) are
    grouping conversations into threads based on subject.

    This returns what seems to be the launchpad way of doing subjects.
    :param  mp: launchpad merge proposal handle
    """
    if '+git' in mp.web_link:
        source = mp.source_git_repository.display_name.replace('lp:', '') + \
                ':' + \
                mp.source_git_path.replace('refs/heads/', '')
        target = mp.target_git_repository.display_name.replace('lp:', '') + \
                ':' + \
                mp.target_git_path.replace('refs/heads/', '')

        return 'Re: [Merge] {} into {}'.format(source, target)
    else:
        return 'Re: [Merge] {} into {}'.format(
            mp.source_branch.display_name,
            mp.target_branch.display_name)


def get_branch_from_mp(merge_proposal):
    """Return a link to branch given a link to a merge proposal.

    If merge_proposal is:
      https://copde.launchpad.net/~user/project/name/+merge/12345
    then the result will be:
      https://copde.launchpad.net/~user/project/name/

    :param merge_proposal: url of a launchpad merge proposal
    """
    m = re.search('(.*)\+merge/[0-9]+$', merge_proposal)
    if m:
        return m.group(1)
    return None


def close_bugs(mp):
    """ Close bugs that are linked to this merge proposal.

    This function goes through all linked bug reports and closes all bugs
    """
    for b in mp.source_branch.linked_bugs_collection:
        for task in b.bug_tasks_collection:
            if task.target.web_link == mp.target_branch.project.web_link:
                task.status = 'Fix Committed'
                task.lp_save()
                message = 'Fix committed into {target} at revision ' + \
                          '{revision}, scheduled for release in ' + \
                          '{project}, milestone {milestone}'
                mp.lp_refresh()
                project = mp.target_branch.project
                milestone_name = 'Unknown'
                active_milestones = project.active_milestones_collection
                for active_milestone in active_milestones:
                    series = active_milestone.series_target
                    if series.branch_link == mp.target_branch.self_link:
                        milestone_name = active_milestone.name
                message = message.format(
                    target=mp.target_branch.display_name,
                    revision=mp.merged_revno,
                    project=project.name,
                    milestone=milestone_name)
                b.newMessage(content=message)


def close_project_bugs(lp_handle, project, commit_msg, bug_comment,
                       bug_status='Fix Committed'):
    """Close any bugs specifically called out in the commit message."""
    bug_ids = []
    for line in commit_msg.splitlines():
        match = re.match(RE_LP_BUG_IDS, line)
        if match:
            bug_ids.extend(
                [re.sub('#|,', '', bug)
                 for bug in match.group().split() if '#' in bug])
    if bug_ids:
        bug_ids = set(bug_ids)
        logger.debug(
            'Setting status of bugs to {0}:\n{1}'.format(
                bug_status,
                '\n'.join(
                    'http://pad.lv/{0}'.format(bug_id) for bug_id in bug_ids
                )
            )
        )

    for bug_id in bug_ids:
        bug = lp_handle.bugs[int(bug_id)]
        bug.newMessage(content=bug_comment)
        for task in bug.bug_tasks:
            if task.bug_target_name == project:
                if task.status != bug_status:
                    task.status = bug_status
                    task.lp_save()


def get_reviews(mp):
    """For a given merge proposal get all reviews.

    Returns list of reviews or None if there are no reviews so far.

    :param mp: merge proposal
    """

    reviews = []
    for vote in mp.votes:
        if vote.comment:
            reviews.append('%s;%s' % (vote.reviewer.display_name,
                                      vote.comment.vote))
    if len(reviews) == 0:
        return None

    return reviews


def get_commit_message(mp, use_description):
    """Get commit message for a given merge proposal.

    If use_description is True and there is no commit message a descritpion
    will be returned instead.

    :param use_description: boolean flag indicating whether we should fallback
                            to description for commit message
    """
    message = None
    if mp.commit_message:
        message = mp.commit_message
    elif use_description and mp.description:
        message = mp.description
    return message


def is_commit_message_set(mp, jenkins_job, jenkins_url):
    """Check if commit message is defined for a given merge proposal.

    This also checks the given jenkins job for an use_description_for_commit
    parameter. If it exists in the jenkins job then the commit message
    fallbacks to description.

    :param mp: launchpad handle to merge proposal
    :param jenkins_job: name of the jenkins job (use_description_for_commit is
                        searched here)
    :param jenkins_url: jenkins url
    """

    jenkins = get_json_jenkins()
    job_url = jenkins_url + '/job/' + jenkins_job + '/'
    use_description = False
    try:
        if jenkinsutils.get_parameter(jenkins, job_url,
                                      'use_description_for_commit'):
            use_description = True
    except KeyError:
        logger.debug("Unable to get \"use_description_for_commit\" " +
                     "configuration for %s" % (jenkins_job))
    logger.debug('Job is configured to fallback to description: %s' %
                 (use_description))
    return bool(get_commit_message(mp, use_description))


def is_git_project(mp):
    """Return boolean if project is a git project."""
    if '+git' in mp.web_link:
        return True
    else:
        return False


def unapproved_prerequisite_exists(mp):
    """Check if there is an unapproved prerequisite branch for a given merge
    proposal.

    :param mp: launchpadlib handle to merge proposal
    """
    prereq = mp.prerequisite_branch
    target_branch = mp.target_branch
    if prereq:
        #x is a merge-proposal object
        merges = [x for x in prereq.landing_targets
                  if x.target_branch.web_link == target_branch.web_link and
                  x.queue_status != u'Superseded']
        if len(merges) == 0:
            logger.debug('No proposals of prerequisite branch exists')
            return True
        else:
            for merge in merges:
                #lets make our life simpler here and lets assume
                #all the MP needs to be merged even though in reality
                #their target branch might differ
                if merge.queue_status != u'Merged':
                    logger.debug('Prerequisite not yet merged')
                    return True
    #TODO
    # skip also branches with changes made after approval
    return False


def testing_in_progress(mp, jenkins_job):
    """ Returns whether testing is in progress.

    This method queries jenkins to find out if we are not running a similar job
    (or downstream such as generic_land) already.

    :param mp: launchpad merge proposal handle
    :param jenkins_job: name of the jenkins job that is going to be triggered
    """
    if jenkinsutils.is_job_or_downstream_building(
            jenkins_job, job_params={'merge_proposal': mp.web_link}):
        logger.debug('Skipping this MP. It is curently being ' +
                     'tested by Jenkins.')
        return True
    return False


def users_in_team(users, team):
    """Determine whether any of these users are in the supplied team.

    :param users: The users which may be members of the supplied team.
    :param team: The team which users may be part of.
    :return: True if any of the users are members of the team, otherwise
        False.
    """
    for member in team.participants:
        if member in users:
            return True
    else:
        return False


def users_allowed_to_trigger_jobs(lp_users):
    """Returns if an of the given users is allowed to run jobs on jenkins.

    This is to avoid random people to run jobs on our internal infrastructure.
    A user is allowed if they are either directly in the ALLOWED_USERS list or
    are member of a team in that list.

    :param lp_users: launchpad user handles
    """
    if len(lp_users) == 0:
        return False
    allowed_people = get_config_option('allowed_users')
    for lp_user in lp_users:
        if lp_user.name in allowed_people:
            return True
    lp = lp_users[0]._root
    for allowed in allowed_people:
        try:
            allowed_person = lp.people[allowed]
        except KeyError:
            logger.warn('User {} from the allowed_users list is not in '
                        'launchpad!'.format(allowed))
            continue
        if not allowed_person.is_team:
            continue
        if users_in_team(lp_users, allowed_person):
            return True
    logger.debug('Users "' + ', '.join(u.name for u in lp_users) +
                 '" not allowed to trigger jobs')
    return False


def is_merge_only(mp):
    """Determines if an mp can be merged without a build"""
    # diffstat is a dictionary of {file: changeset}
    # diffstat will be None if launchpad has not yet generated a diff
    if mp.preview_diff is None or mp.preview_diff.diffstat is None:
        raise LookupError("preview_diff of this merge proposal is None")
    diff_stat = mp.preview_diff.diffstat
    logger.debug("%s" % (diff_stat))
    if (len(diff_stat) == 1 and u'debian/changelog' in diff_stat):
        logger.debug("debian/changelog only merge")
        return True
    return False

def get_review_revision_regex(mp):
    if '+git' in mp.web_link:
        return '^(PASSED|FAILED): Continuous integration, rev:([0-9a-f]+)'
    else:
        return '^(PASSED|FAILED): Continuous integration, rev:(\d+)'

def get_latest_review(launchpad_user, mp):
    """Return the latest revision reviewed by the given launchpad_user.

    This function expects a review comment in the following format:
       '^(PASSED|FAILED): Continuous integration, rev:(\d+)'

    :param launchpad_user: handle to launchpad user
    :param mp: handle to merge proposal
    """
    revision = 0
    launchpad_review_type = get_config_option('launchpad_review_type')
    for comment in mp.all_comments:
        if comment.author.name == launchpad_user.name:
            if comment.vote_tag == launchpad_review_type:
                m = re.search(
                    get_review_revision_regex(mp),
                    comment.message_body)
                if m:
                    revision = m.group(2)
    return revision

def get_latest_revision(mp):
    """Return the latest revision of the given merge proposal.

    :param mp: handle to merge proposal
    """
    if '+git' in mp.web_link:
        for ref in mp.source_git_repository.refs_collection:
            if ref.path == mp.source_git_path:
                return ref.commit_sha1
        return str(0)
    else:
        return mp.source_branch.revision_count


def latest_candidate_validated(launchpad_user, mp):
    """Return if the latest candidate revision of the merge proposal is
    validated.

    :param launchpad_user: handle to launchpad user used to validate the
                           merge proposals
    :param mp: handle to merge proposal
    """

    latest_review = get_latest_review(launchpad_user, mp)
    logger.debug('Latest review is revision: ' + str(latest_review))
    latest_revision = get_latest_revision(mp)
    logger.debug('Latest revision is: ' + str(latest_revision))
    if str(latest_review) == str(latest_revision):
        logger.debug('Skipping this MP. Current revision: ' +
                     str(latest_revision))
        return True
    return False


def report_to_launchpad(mp, message, revision, status=None, vote=None):
    """Set the status and send a message to Launchpad for an merge proposal.

    :param mp: handle to merge proposal
    :param message: text message to post as a message
    :param revision: revision number for the proposal
    :param status: Launchpad status to set the propsal to
    :param vote: how the message should vote on the proposal
    """
    if status:
        mp.setStatus(status=status, revid=revision)
    if vote:
        mp.createComment(
            review_type=get_config_option('launchpad_review_type'),
            vote=vote,
            subject=get_vote_subject(mp),
            content=message)
    else:
        mp.createComment(
            review_type=get_config_option('launchpad_review_type'),
            subject=get_vote_subject(mp),
            content=message)


def get_template_args(result, revision, build_url, merge_url):
    """Collect a set of information for rendering templates

    :param result: Result of the tests, either PASSED or FAILED
    :param revison: The revision number of the merge
    :param build_url: URL to the build triggered in jenkins
    :param merge_url: URL to the merge proposal
    """
    jenkins = get_json_jenkins()
    args = {}
    args['result'] = result
    args['revision'] = revision
    args['build_url'] = build_url
    args['merge_url'] = merge_url
    build_urls = jenkinsutils.get_executed_builds(jenkins, build_url)
    args['coverity_artifacts'] = []
    args['executed_tests'] = []
    for build in build_urls:
        if build['output'].endswith('console'):
            build_url = build['output'][:-len('console')]
            build['output'] = build_url
            debs = jenkinsutils.get_deb_artifacts(jenkins, build_url)
            if debs is None:
                build['deb'] = []
            else:
                build['deb'] = debs
            args['coverity_artifacts'].extend(
                jenkinsutils.get_coverity_artifacts(jenkins, build_url)
            )
        args['executed_tests'].append(build)
    return args


class UpdateMergeProposalException(Exception):
    pass


def get_branch_handle_for(lp_handle, name, repo_type='auto'):
    """ Return the branch type and repo handle for the given name.

    Returns the repo_type and a launchpad branch or git repository
      handle for the given name.
    :param lp_handle: launchpad API handle/instance
    :param name: name of the branch or git repository
    :param repo_type: Type of repository [bazaar, git, or auto]
    """
    logger.debug('Provided branch name: {}'.format(name))
    logger.debug('Using repo_type: {}'.format(repo_type))

    branch = lp_handle.branches.getByUrl(url=name)
    if repo_type == 'bazaar':
        logger.debug('bazaar repo: {}'.format(branch))
        return repo_type, branch

    try:
        repo = lp_handle.git_repositories.getByPath(
            path=name.replace('lp:', ''))
    except AttributeError:
        logger.debug('git_repositories.getByPath was not found. You may need '
                     'to set lp_version:devel in the config')
        if repo_type == 'git':
            raise RuntimeError('Git support is incomplete, '
                               'set lp_version:devel in the config.')

    if repo_type == 'git':
        logger.debug('git repo: {}'.format(repo))
        return repo_type, repo
    elif repo_type == 'auto':
        if branch is not None and repo is not None:
            # Ambiguous name, can't automatically select
            raise RuntimeError('Ambiguous branch name: {} resolves to '
                               '{} and {}'.format(name, branch, repo))
        if branch is not None:
            logger.debug('bazaar repo: {}'.format(branch))
            return 'bazaar', branch
        elif repo is not None:
            logger.debug('git repo: {}'.format(repo))
            return 'git', repo
        else:
            raise RuntimeError('Could not determine branch or repo from '
                               '{}'.format(name))
    else:
        raise RuntimeError('Repo type "' + repo_type + '" is not valid')


def get_merge_proposals(lp_handle, name, state, repo_type='auto'):
    """ Return all merge proposals for a given branch.

    Returns a list of launchpad handles for merge proposals.
    :param lp_handle: launchpad API handle/instance
    :param name: name of the branch or git repository
    :param state: list of states I'm interested in (usually something like
    ['Approved'] or ['Needs review'])
    :param repo_type: Type of repository [bazaar, git, or auto]
    """

    _, branch = get_branch_handle_for(lp_handle, name, repo_type)
    if not branch:
        logger.debug('Branch "' + name + '" doesn\'t exist')
        return []

    mps = []
    for mp in branch.landing_candidates:
        if mp.queue_status in state:
            mps.append(mp)
    return mps

def get_branch_handle_from_url(lp_handle, url):
    """ Return a branch/repo handle for the given url.

    Returns a launchpad branch or git repository handle for the given url.
    :param lp_handle: launchpad API handle/instance
    :param url: url of the branch or git repository
    """
    if '+git' in url:
        name = url.replace('https://code.launchpad.net/', '')
        logger.debug('fetching repo: ' + name)
        try:
           return lp_handle.git_repositories.getByPath(path=name)
        except AttributeError:
            logger.debug('git_repositories.getByPath was not found. You may need to set lp_version=devel in the config')
            return None
    else:
        name = url.replace('https://code.launchpad.net/', 'lp:')
        name = name.replace('https://code.staging.launchpad.net/', 'lp://staging/')
        logger.debug('fetching branch: ' + name)
        return lp_handle.branches.getByUrl(url=name)

def get_mp_handle_from_url(lp_handle, merge_proposal_link):
    """ Get launchpad handle for merge proposal given a merge proposal URL.

    Returns None in case the merge proposal can't be found.
    :param merge_proposal_link: URL of the merge proposal
    """
    branch_link = get_branch_from_mp(merge_proposal_link)
    if not branch_link:
        logger.error('Unable to get branch link from merge proposal link.')
        return None

    branch = get_branch_handle_from_url(lp_handle, branch_link)
    if not branch:
        logger.debug('Branch {} does not exist'.format(branch_link))
        return None

    logger.debug('mp_link: {}.'.format(merge_proposal_link))

    for mp in branch.landing_targets:
        logger.debug('mp.web_link: {}'.format(mp.web_link))
        if mp.web_link == merge_proposal_link:
            return mp

    return None
