import unittest
from mock import patch, MagicMock
import tests.util as util
from jlp.commands.voteOnMergeProposal import voteOnMergeProposal


class TestVoteOnMergeProposal(unittest.TestCase):

    def setUp(self):
        get_launchpad = lambda launchpadlib_dir: MagicMock()
        self.get_launchpad_patch = patch(
            'jlp.commands.voteOnMergeProposal.get_launchpad',
            new=get_launchpad)
        self.get_launchpad_patch.start()
        self.is_commit_message_set_patch = patch(
            'jlp.launchpadutils.is_commit_message_set')
        self.is_commit_message_set_patch.start()

    def tearDown(self):
        self.get_launchpad_patch.stop()
        self.is_commit_message_set_patch.stop()

    def test_gibberish(self):
        """Gibberish on the command line produces usage message."""

        sys_argv = ['--oheck']
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            try:
                voteOnMergeProposal()
            except SystemExit:
                pass
            self.assertTrue("usage" in stderr.getvalue())

    def test_no_build_url_specified(self):
        """No build_url specified; exit with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal="https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113"']
        error_message = 'error: argument -u/--build-url is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                voteOnMergeProposal()
            self.assertTrue(error_message in stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_no_revision_specified(self):
        """No revision specified; exit with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url="http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure"',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal="https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113"']
        error_message = 'error: argument -r/--revision is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                voteOnMergeProposal()
            self.assertTrue(error_message in stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_no_merge_proposal_specified(self):
        """No merge_proposal specified; exit with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url="http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure"',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"']
        error_message = 'error: argument -p/--merge-proposal is required'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr:
            with self.assertRaises(SystemExit) as e:
                voteOnMergeProposal()
            self.assertTrue(error_message in stderr.getvalue())
            self.assertEquals(2, e.exception.code)

    def test_unknown_merge_proposal(self):
        """The given merge proposal wasn't found; error out with message."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url="http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure"',
                    '--revision=66',
                    '--branch="lp:~mrazik/faux-dbus-test-runner/yar8"',
                    '--merge-proposal=https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113']
        error_message = 'error: merge proposal related to this branch was ' + \
                        'not found'
        with patch('sys.argv', sys_argv), \
                util.captured_stderr() as stderr, \
                patch('jlp.commands.voteOnMergeProposal.launchpadutils.' +
                      'get_mp_handle_from_url') as get_mp_handle_from_url:
            get_mp_handle_from_url.return_value = None
            try:
                voteOnMergeProposal()
            except SystemExit:
                pass
            self.assertTrue(error_message in stderr.getvalue())

    @patch('sys.argv')
    @patch('jlp.commands.voteOnMergeProposal.launchpadutils.get_mp_handle_from_url')
    @patch('jlp.launchpad.report_to_launchpad')
    @patch('jlp.commands.voteOnMergeProposal.launchpadutils.get_template_args')
    @patch('jlp.commands.voteOnMergeProposal.get_jinja_environment')
    def test_passed_vote(self, gje, gta, rtl, gmphandle, sys_argv):
        """Vote on a proposal which passed."""

        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url=http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure',
                    '--revision=66',
                    '--branch=lp:~mrazik/faux-dbus-test-runner/yar8',
                    '--merge-proposal=https://code.launchpad.net/~mrazik/' +
                    'faux-dbus-test-runner/yar8/+merge/106113']

        gmphandle.return_value = 'Launchpad_mp'
        gta.return_value = {
            'result': 'PASSED',
            'revision': '66'
        }
        ret = voteOnMergeProposal()
        environment = MagicMock
        gje.return_value = environment
        gt_mock = MagicMock
        environment.get_template.return_value = gt_mock
        gt_mock.return_value = MagicMock
        self.assertEqual(ret, 0)
        rtl.assert_called_once()
        gt_mock.assert_called_with('ci_test.j2')


    @patch('sys.argv')
    @patch('jlp.commands.voteOnMergeProposal.launchpadutils.get_mp_handle_from_url')
    @patch('jlp.launchpad.report_to_launchpad')
    @patch('jlp.launchpadutils.is_commit_message_set')
    @patch('jlp.commands.voteOnMergeProposal.launchpadutils.get_template_args')
    @patch('jlp.commands.voteOnMergeProposal.get_jinja_environment')
    def test_no_commit_message_set(self, gje, gta, icms, rtl, gmphandle, sys_argv):
        """Vote on a proposal where tests passed but no commit message
           was set."""

        merge_proposal = 'https://code.launchpad.net/~mrazik/' + \
                         'faux-dbus-test-runner/yar8/+merge/106113'
        sys_argv = ['voteOnMergeProposal.py',
                    '--status=PASSED',
                    '--build-url=http://s-jenkins.ubuntu-ci/job/' +
                    'faux-dbus-test-runner-ci/configure',
                    '--revision=66',
                    '--branch=lp:~mrazik/faux-dbus-test-runner/yar8',
                    '--merge-proposal=' + merge_proposal]

        gmphandle.return_value = 'Launchpad_mp'
        gta.return_value = {
            'result': 'PASSED',
            'revision': '66',
            'build_url': 'http://s-jenkins.ubuntu-ci/job/' +
                         'faux-dbus-test-runner-ci/configure'
        }
        icms.return_value = False
        environment = MagicMock
        gje.return_value = environment
        gt_mock = MagicMock
        environment.get_template.return_value = gt_mock
        gt_mock.return_value = MagicMock
        
        ret = voteOnMergeProposal()
        self.assertEqual(ret, 0)
        gt_mock.assert_called_with('no_commit.j2')
